"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.queue = exports.tick = exports.last = void 0;
function last(things) {
    return things.length ? things[things.length - 1] : undefined;
}
exports.last = last;
function tick(task) {
    process.nextTick(task);
}
exports.tick = tick;
function queue() {
    const taskQueue = [];
    let ongoing = false;
    let callbacks = [];
    const execute = () => {
        const next = taskQueue.shift();
        if (next !== undefined) {
            try {
                next();
            }
            finally {
                process.nextTick(execute);
            }
        }
        else {
            ongoing = false;
            callbacks.forEach((task) => process.nextTick(task));
            callbacks = [];
        }
    };
    return {
        push(task) {
            taskQueue.push(task);
            if (!ongoing) {
                ongoing = true;
                process.nextTick(execute);
            }
        },
        onDone(task) {
            if (ongoing) {
                callbacks.push(task);
            }
            else {
                process.nextTick(task);
            }
        },
    };
}
exports.queue = queue;
